/*
 * The Gold Partition Conjecture for K-thin posets
 *
 * gpc.h
 * General definitions
 */

/* Copyright 2005, Marcin Peczarski. All Rights Reserved.
 *
 * This program is free software. You may use this software, and copy,
 * redistribute, and/or modify it for non profit works.
 * Please do not remove this copyright notice.
 */

#ifndef _GPC_H_
#define _GPC_H_

/*
 * Compilation control constants (currently defined in makefile)
 *
 * Architecture definition:
 *   _IA32_                -- Linux on 32-bit x86 machine
 *   _X86_64_              -- Linux on 64-bit x86 machine
 *
 * The experiment definition:
 *   K                     -- thinness
 *
 * Other possible definition:
 *   _INT_INDEX_           -- int indices in hash table, use carefully if
 *                            MAX_STACK * sizeof(POSET) > INT_MAX
 *   _PRINT_LOG_           -- print log to standard output
 *   _CHECK_CORRECTNESS_   -- include some checking code
 */

#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

/*
 * Experiment definition
 */

#if K <= 4
  #define N                    12
  #define FULL_SET         0xFFFU
  #define MAX_STACK           180
  #define HASH_TBL_SIZE       227
  #define MAX_COMPONENTS       24
  #define READ_INTERVAL       100
#elif K == 5
  #define N                    22
  #define FULL_SET      0x3FFFFFU
  #define MAX_STACK          1188
  #define HASH_TBL_SIZE      1487
  #define MAX_COMPONENTS      164
  #define READ_INTERVAL      1000
#elif K == 6
  #define N                    32
  #define FULL_SET    0xFFFFFFFFU
  #define MAX_STACK         27000
  #define HASH_TBL_SIZE     36083
  #define MAX_COMPONENTS     1100
  #define READ_INTERVAL      9000
#else
  #define N                    32
  #define FULL_SET    0xFFFFFFFFU
  #define MAX_STACK        200000
  #define HASH_TBL_SIZE    200087
  #define MAX_COMPONENTS     1600
  #define READ_INTERVAL      9000
#endif

/*
 * Common types, functions and macros
 */

#define EMPTY_SET 0

typedef int BOOL;
#define TRUE  1
#define FALSE 0

#if defined(_IA32_) && N <= 12
  typedef int CPC;
  #define CPC_PRINT_FORMAT "%d"
#elif defined(_IA32_) && N <= 18
  typedef double CPC;
  #define CPC_PRINT_FORMAT "%-.0f"
#elif defined(_IA32_)
  typedef long double CPC;
  #define CPC_PRINT_FORMAT "%-.0Lf"
#elif defined(_X86_86_) && N <= 12
  typedef int CPC;
  #define CPC_PRINT_FORMAT "%d"
#elif defined(_X86_64_)
  typedef long CPC;
  #define CPC_PRINT_FORMAT "%ld"
#elif defined(_ILP32_) && N <= 12
  typedef int CPC;
  #define CPC_PRINT_FORMAT "%d"
#elif defined(_ILP32_) && N <= 20
  typedef long long CPC;
  #define CPC_PRINT_FORMAT "%Ld"
#elif defined(_LP64_) && N <= 12
  typedef int CPC;
  #define CPC_PRINT_FORMAT "%d"
#elif defined(_LP64_) && N <= 20
  typedef long CPC;
  #define CPC_PRINT_FORMAT "%ld"
#else
  #error The number of linear extensions cannot be counted for this N.
#endif

#if N <= 16
  typedef unsigned short SET;
#elif N <= 32
  typedef unsigned int SET;
#elif N <= 64 && (defined(_X86_64_) || defined(_LP64_))
  typedef unsigned long SET;
#elif N <= 64 && (defined(_IA32_) || defined(_ILP32_))
  typedef unsigned long long SET;
#else
  #error N is too large.
#endif

#ifdef _INT_INDEX_
  typedef int INDEX;
#else
  typedef long INDEX;
#endif

#if defined(_IA32_) || defined(_X86_64_)
  #define _USE_POF2_TABLE_
  #if defined(__GNUC__) && !defined(__STRICT_ANSI__)
    #define _USE_BUILT_IN_MIN_ELEMENT_
  #else
    #define _USE_MIN_ELEMENT_TABLE_
    typedef char MIN_EL_TBL_EL;
  #endif
  #define _USE_CARDINALITY_TABLE_
  typedef char CARD_TBL_EL;
#else
  #define _USE_MIN_ELEMENT_TABLE_
  typedef int MIN_EL_TBL_EL;
  #define _USE_CARDINALITY_TABLE_
  typedef int CARD_TBL_EL;
#endif

#ifdef _USE_POF2_TABLE_
  /* To look in a table is faster than shift operation! */
  #define pof2(x) pof2_tbl[x]
  #define in(x, n) (((x) & pof2_tbl[n]) != EMPTY_SET)
  #define notin(x, n) (((x) & pof2_tbl[n]) == EMPTY_SET)
#else
  #define pof2(x) (1 << (x))
  #define in(x, n) (((x) & (1 << (n))) != EMPTY_SET)
  #define notin(x, n) (((x) & (1 << (n))) == EMPTY_SET)
#endif

#if defined(_USE_MIN_ELEMENT_TABLE_)
  #if N <= 24
    #define MET_SIZE       (1 << N)
    #define min_element(x) (int)min_element_tbl[x]
  #else
    #define MET_HALF_N     ((N + 1) / 2)
    #define MET_SIZE       (1 << MET_HALF_N)
    #define MET_HALF_SET   (FULL_SET >> (N / 2))
    #define min_element(x) ((x) & MET_HALF_SET ?                             \
                            min_element_tbl[(x) & MET_HALF_SET] :            \
                            min_element_tbl[(x) >> MET_HALF_N] + MET_HALF_N)
  #endif
#elif defined(_USE_BUILT_IN_MIN_ELEMENT_)
  #if defined(_IA32_) && N <= 32
    #define min_element(set) ({                                              \
      register int element;                                                  \
      asm ("bsfl %1, %0" : "=r"(element) : "g"((unsigned int)(set)) : "cc"); \
      element;                                                               \
    })
  #elif defined(_X86_64_) && N <= 16
    #define min_element(set) ({                                     \
      register short element;                                       \
      asm ("bsfw %1, %0" : "=r"(element) : "g"((SET)(set)) : "cc"); \
      element;                                                      \
    })
  #elif defined(_X86_64_) && N <= 32
    #define min_element(set) ({                                     \
      register int element;                                         \
      asm ("bsfl %1, %0" : "=r"(element) : "g"((SET)(set)) : "cc"); \
      element;                                                      \
    })
  #elif defined(_X86_64_) && N <= 64
    #define min_element(set) ({                                     \
      register long element;                                        \
      asm ("bsfq %1, %0" : "=r"(element) : "g"((SET)(set)) : "cc"); \
      element;                                                      \
    })
  #else
    #error There is no built-in min_element function.
  #endif
#else
  int min_element(SET);
#endif

#if defined(_USE_CARDINALITY_TABLE_)
  #if N <= 24
    #define CAT_SIZE       (1 << N)
    #define cardinality(x) (int)cardinality_tbl[x]
  #else
    #define CAT_HALF_N     ((N + 1) / 2)
    #define CAT_SIZE       (1 << CAT_HALF_N)
    #define CAT_HALF_SET   (FULL_SET >> (N / 2))
    #define cardinality(x) (cardinality_tbl[(x) & CAT_HALF_SET] + \
                            cardinality_tbl[(x) >> CAT_HALF_N])
  #endif
#elif defined(_USE_BUILT_IN_CARDINALITY_)
  #error There is not built-in cardinality function.
#else
  int cardinality(SET);
#endif

#if N <= 33
  /* 32-bit integer is sufficient and faster. */
  typedef int BINOM;
#else
  typedef CPC BINOM;
#endif

typedef struct {
  INDEX hash;
  int   m, n;
  SET   r[2][N]; /* r[0][j] - elements above j, r[1][j] - elements below j */
  SET   M;
} POSET;

#define poset_empty(g) memset(g, 0, sizeof(POSET))

typedef struct {
  int    thinness;
  int    version;
  int    step;
  int    max_elements;
  long   input_offset;
  long   input_posets;
  long   output_posets;
  long   considered_posets;
  long   isomorphism_calls;
  long   combinations;
  long   hash_tbl_usage;
  long   stack_usage;
  time_t time;
  CPC    max_extensions;
  int    max_components;
} FILE_HEADER;

#define FATR 0644 /* atributes for a new file */
#define VERSION (100000*N + 100*sizeof(POSET) + sizeof(FILE_HEADER))

void   gpc2_module_init(void);
void   gpc2_module_clean(void);
void   poset_module_init(void);
void   poset_module_clean(void);
void   fatal_error(const char*, ...);
void   poset_print(const POSET*);
BOOL   isomorphic(const POSET*, const POSET*);
void   reset_read_buffer(long);
POSET* get_next_poset(void);
int    first_configuration(void);
int    one_step(BOOL, long, long);
void   print_file_header(const FILE_HEADER*);
void   empty_file_header(FILE_HEADER*);
CPC    cpcp_i(SET, const SET*, const SET*);

#define cpcp(g) cpcp_i((g)->M, (g)->r[0], (g)->r[1])

#ifdef _POSET_C_
  #ifdef _USE_POF2_TABLE_
  /* One element set - Power OF 2 table */
  SET pof2_tbl[N];
  #endif
  #ifdef _USE_MIN_ELEMENT_TABLE_
  /* Lowest element in the set */
  MIN_EL_TBL_EL min_element_tbl[MET_SIZE + 384];
  #endif
  #ifdef _USE_CARDINALITY_TABLE_
  /* Set cardinality - counting ones */
  CARD_TBL_EL cardinality_tbl[CAT_SIZE + 384];
  #endif
#else
  #ifdef _USE_POF2_TABLE_
  /* One element set - Power OF 2 table */
  extern SET pof2_tbl[];
  #endif
  #ifdef _USE_MIN_ELEMENT_TABLE_
  /* Lowest element in the set */
  extern MIN_EL_TBL_EL min_element_tbl[];
  #endif
  #ifdef _USE_CARDINALITY_TABLE_
  /* Set cardinality - counting ones */
  extern CARD_TBL_EL cardinality_tbl[];
  #endif
#endif

#endif /* _GPC_H_ */
