/*
 * The Gold Partition Conjecture for K-thin posets
 *
 * gpc1.c
 * Main module
 */

/* Copyright 2005, Marcin Peczarski. All Rights Reserved.
 *
 * This program is free software. You may use this software, and copy,
 * redistribute, and/or modify it for non profit works.
 * Please do not remove this copyright notice.
 */

#include "gpc.h"

static void main_init(void);
static void main_clean(void);
static void print_files_contents(int, char**, BOOL);

int in_file = -1;
int out_file = -1;
time_t start_time;

/*
 * Global data initialization and deinitialization
 */

static void main_init(void) {
  gpc2_module_init();
  poset_module_init();
}

static void main_clean(void) {
  if (in_file >= 0)
    close(in_file);
  if (out_file >= 0)
    close(out_file);
  gpc2_module_clean();
  poset_module_clean();
}

/*
 * Program administration
 */

void fatal_error(const char *fmt, ...) {
  va_list fmt_args;

  fprintf(stderr, "ERROR: ");
  va_start(fmt_args, fmt);
  vfprintf(stderr, fmt, fmt_args);
  va_end(fmt_args);
  if (errno) {
    fprintf(stderr, ", errno=%d", errno);
    perror(" ");
  }
  else
    fprintf(stderr, "\n");
  main_clean();
  exit(2);
}

int main(int argc, char *argv[]) {
  long offset, count;
  int  action, result = 0;

  if (argc < 2) {
    printf("usage: %s action [parameters]\ncompiled at "__DATE__
           " "__TIME__"\n", argv[0]);
    #ifdef _INT_INDEX_
    printf("int indices\n");
    #else
    printf("long indices\n");
    #endif
    #ifdef _PRINT_LOG_
    printf("printing log on\n");
    #else
    printf("printing log off\n");
    #endif
    #ifdef _CHECK_CORRECTNESS_
    printf("checking code included\n");
    #else
    printf("checking code not included\n");
    #endif
    #if defined _USE_MIN_ELEMENT_TABLE_
      printf("min element table\n");
    #elif defined _USE_BUILT_IN_MIN_ELEMENT_
      printf("built-in min element\n");
    #endif
    #if defined _USE_BUILT_IN_MAX_ELEMENT_
      printf("built-in max element\n");
    #endif
    printf("THINNES = %d   VERSION = %d\n"
           "MAX_STACK = %ld   HASH_TBL_SIZE = %ld   MAX_COMPONENTS = %d\n"
           "READ_INTERVAL = %d\n\n",
           (int)K, (int)VERSION,
           (long)MAX_STACK, (long)HASH_TBL_SIZE, (int)MAX_COMPONENTS,
           (int)READ_INTERVAL);
    return 3;
  }
  start_time = time(NULL);
  action = atoi(argv[1]);
  switch (action) {

    case 0: /* Create initial file containing the first configuration. */
      if (argc < 3) {
        printf("usage: %s 0 out_file\n\n", argv[0]);
        return 3;
      }
      out_file = open(argv[2], O_RDWR | O_CREAT | O_TRUNC, FATR);
      if (out_file < 0)
        fatal_error("can't create file %s", argv[2]);
      first_configuration();
      return 0;

    case 1: /* Make one step. */
      if (argc < 4) {
        printf("usage: %s 1 in_file out_file [offset [count]]\n\n",
               argv[0]);
        return 3;
      }
      offset = argc >= 5 ? atol(argv[4]) : 0;
      count  = argc >= 6 ? atol(argv[5]) : 0;
      in_file = open(argv[2], O_RDONLY);
      if (in_file < 0)
        fatal_error("can't open file %s", argv[2]);
      out_file = open(argv[3], O_RDWR | O_CREAT | O_EXCL, FATR);
      if (out_file < 0 && errno == EEXIST) {
        out_file = open(argv[3], O_RDWR);
        if (out_file < 0)
          fatal_error("can't open file %s", argv[3]);
        else {
          errno = 0;
          main_init();
          result = one_step(TRUE, offset, count);
        }
      }
      else if (out_file < 0)
        fatal_error("can't create file %s", argv[3]);
      else {
        main_init();
        result = one_step(FALSE, offset, count);
      }
      main_clean();
      return result;

    case 5: /* Display overview of a file. */
    case 6: /* Display contents of a file. */
      if (argc < 3) {
        printf("usage: %s %d files_list\n\n", argv[0], action);
        return 3;
      }
      main_init();
      print_files_contents(argc, argv, action == 6);
      main_clean();
      return 0;

    default:
      printf("unknown action: %s\n", argv[1]);
      return 3;
  }
}

/*
 * Auxiliary routines
 */

void empty_file_header(FILE_HEADER* fh) {
  memset(fh, 0, sizeof(FILE_HEADER));
  fh->thinness          = K;
  fh->version           = VERSION;
  fh->step              = 0;
  fh->max_elements      = 0;
  fh->input_offset      = 0;
  fh->input_posets      = 0;
  fh->output_posets     = 0;
  fh->considered_posets = 0;
  fh->isomorphism_calls = 0;
  fh->combinations      = 0;
  fh->hash_tbl_usage    = 0;
  fh->stack_usage       = 0;
  fh->time              = 0;
  fh->max_extensions    = 0;
  fh->max_components    = 0;
}

void compare_file_headers(FILE_HEADER* fh0, const FILE_HEADER* fh1) {
  if (fh0->step < 0)
    fh0->step = fh1->step;
  else if (fh0->step != fh1->step)
    fatal_error("compare_file_headers, files belong to different iterations");
  if (fh0->input_offset   > fh1->input_offset)
      fh0->input_offset   = fh1->input_offset;
  if (fh0->max_elements   < fh1->max_elements)
      fh0->max_elements   = fh1->max_elements;
  if (fh0->hash_tbl_usage < fh1->hash_tbl_usage)
      fh0->hash_tbl_usage = fh1->hash_tbl_usage;
  if (fh0->stack_usage    < fh1->stack_usage)
      fh0->stack_usage    = fh1->stack_usage;
  if (fh0->max_extensions < fh1->max_extensions)
      fh0->max_extensions = fh1->max_extensions;
  if (fh0->max_components  < fh1->max_components)
      fh0->max_components  = fh1->max_components;
  fh0->input_posets      += fh1->input_posets;
  fh0->output_posets     += fh1->output_posets;
  fh0->considered_posets += fh1->considered_posets;
  fh0->isomorphism_calls += fh1->isomorphism_calls;
  fh0->combinations      += fh1->combinations;
  fh0->time              += fh1->time;
}

static void print_files_contents(int argc, char* argv[], BOOL print_posets) {
  FILE_HEADER fh0, fh;
  POSET       *p;
  long        i;
  int         k;

  empty_file_header(&fh0);
  fh0.step = -1;
  fh0.input_offset = (long)((unsigned long)(-1) >> 1);
  for (k = 2; k < argc; k++) {
    in_file = open(argv[k], O_RDONLY);
    if (in_file < 0)
      fatal_error("print_files_contents, can't open file %s", argv[2]);
    if (read(in_file, &fh, sizeof(FILE_HEADER)) != sizeof(FILE_HEADER))
      fatal_error("print_files_contents, read file header failed");
    if (fh.thinness != K)
      fatal_error("print_files_contents, missing thinness %d", fh.thinness);
    if (fh.version != VERSION)
      fatal_error("print_files_contents, missing version %d", fh.version);
    if (print_posets) {
      reset_read_buffer(0);
      for (i = 0; i < fh.output_posets; i++) {
        p = get_next_poset();
        printf("poset %ld  ", i);
        poset_print(p);
      }
    }
    close(in_file);
    compare_file_headers(&fh0, &fh);
  }
  in_file = -1;
  print_file_header(&fh0);
}

void print_file_header(const FILE_HEADER* fh) {
  char cpc[24];

  sprintf(cpc, CPC_PRINT_FORMAT, fh->max_extensions);
  printf("thinness %d  iteration %d  computation time %ld\n"
         "input offset              %12ld\n"
         "input configurations      %12ld\n"
         "output configurations     %12ld\n"
         "considered configurations %12ld\n"
         "isomorphism calls         %12ld\n"
         "combinations checked      %12ld\n"
         "hash table usage          %12ld\n"
         "stack usage               %12ld\n"
         "max elements              %12d\n"
         "max components            %12d\n"
         "max extensions %23.23s\n",
         K, fh->step, fh->time, fh->input_offset,
         fh->input_posets, fh->output_posets, fh->considered_posets,
         fh->isomorphism_calls, fh->combinations, fh->hash_tbl_usage,
         fh->stack_usage, fh->max_elements, fh->max_components, cpc);
}
