/*
 * The Gold Partition Conjecture for K-thin posets
 *
 * poset.c
 * Poset auxiliary routines, counting linear extensions
 */

/* Copyright 2005, Marcin Peczarski. All Rights Reserved.
 *
 * This program is free software. You may use this software, and copy,
 * redistribute, and/or modify it for non profit works.
 * Please do not remove this copyright notice.
 */

#define _POSET_C_
#include "gpc.h"

/* Binomial coefficients: binomial[k][n] = n!/(k!*(n-k)!)  */
static BINOM binomial[N + 1][N + 1];

/* Linear extensions count for small posets */
#define NU -1000 /* NOT USED indicator */
static const int lec3[4] = {6, 3, 2, 1};
static const int theta4[4] = {0, 1, 4, 10};
static const int lec4[16] = {24, 12, 6, 6, 8,  5, 3, NU,
                              4,  2, 6, 3, 2, NU, 2,  1};
static const int theta5[5][5] = {{NU,  0,  1,  3, 10},
                                 { 0,  6,  8,  5, NU},
                                 { 1,  8,  1, NU, NU},
                                 { 3,  5, NU, NU, NU},
                                 {10, NU, NU, NU, NU}};
static const int lec5[43] = {NU, NU, NU, 16, 18, NU, 14, NU, NU, 12,
                             24,  9, 11,  4,  8,  7, NU, 12,  6, NU,
                              8, NU,  2, NU,  6,  8, NU,  3,  5,  4,
                             NU,  1,  4, NU, NU, NU, NU,  NU, 6, NU,
                             NU,  2,  3};

void poset_module_init(void) {
  int i, j;

  #ifdef _USE_POF2_TABLE_
  SET p;

  for (i = 0, p = 1; i < N; i++, p <<= 1)
    pof2_tbl[i] = p;
  #endif

  #ifdef _USE_MIN_ELEMENT_TABLE_
  min_element_tbl[0] = 0;
  min_element_tbl[1] = 0;
  i = 1;
  while (i < MET_SIZE) {
    i++;
    /* i is even */
    min_element_tbl[i] = min_element_tbl[i >> 1] + 1;
    i++;
    /* i is odd */
    min_element_tbl[i] = 0;
  }
  #endif

  #ifdef _USE_CARDINALITY_TABLE_
  cardinality_tbl[0] = 0;
  cardinality_tbl[1] = 1;
  i = 1;
  while (i < CAT_SIZE) {
    i++;
    /* i is even */
    cardinality_tbl[i] = cardinality_tbl[i >> 1];
    i++;
    /* i is odd */
    cardinality_tbl[i] = cardinality_tbl[i >> 1] + 1;
  }
  #endif

  binomial[0][0] = 1;
  for (i = 1; i <= N; i++) {
    binomial[0][i] = 1;
    for (j = 1; j < i; j++)
      binomial[j][i] = binomial[j - 1][i - 1] + binomial[j][i - 1];
    binomial[i][i] = 1;
  }
}

void poset_module_clean(void) {
}

void poset_print(const POSET *g) {
  int i, j, k;
  SET V;
  static const char description[2][6] = {"above", "below"};

  printf("m=%d  n=%d  M=0x%X\n", g->m, g->n, g->M);
  for (k = 0; k < 2; k++) {
    printf("%s\n", description[k]);
    for (i = 0; i < g->n; i++) {
      V = g->r[k][i];
      if (V) {
        printf("%d:", i);
        do {
          j = min_element(V);
          V ^= pof2(j);
          printf(" %d", j);
        } while(V);
        printf("\n");
      }
    }
  }
}

/*
 * Linear extensions counting funcion
 */

CPC cpcp_i(SET component_set, const SET *ra, const SET *rb) {
  int c, k, v, split_v, w, x, y, count, max_count;
  SET component, prev_component, new_elements, v_set;
  SET m_set, m_sub_set, sub_component;
  CPC t, t1, t2;

  t = 1;
  c = 0;
  while (component_set) {
    /* Calculate next connected component of component_set,
       then substract it from component_set. */
    v = min_element(component_set);
    prev_component = pof2(v);
    component = (prev_component | ra[v] | rb[v]) & component_set;
    while (component != prev_component) {
      new_elements = component & ~prev_component;
      prev_component = component;
      while (new_elements) {
        v = min_element(new_elements);
        new_elements ^= pof2(v);
        component |= ra[v] | rb[v];
      }
      component &= component_set;
    }
    component_set &= ~component;

    k = cardinality(component);
    c += k;
    switch (k) {
      case 1: /* single element component */
        t *= c;
        break;
      case 2: /* 2-element connected component */
        t *= binomial[2][c];
        break;
      case 3: /* 3-element connected component */
        v = min_element(component);
        v_set = component ^ pof2(v);
        w = min_element(v_set);
        t *= lec3[cardinality(ra[v] & component) +
                  cardinality(ra[w] & component) +
                  cardinality(ra[min_element(v_set ^ pof2(w))] & component)] *
             binomial[3][c];
        break;
      case 4: /* 4-element connected component */
        v = min_element(component);
        v_set = component ^ pof2(v);
        w = min_element(v_set);
        v_set ^= pof2(w);
        x = min_element(v_set);
        t *= lec4[theta4[cardinality(ra[v] & component)] +
                  theta4[cardinality(ra[w] & component)] +
                  theta4[cardinality(ra[x] & component)] +
                  theta4[cardinality(ra[min_element(v_set ^ pof2(x))] &
                                     component)]] * binomial[4][c];
        break;
      case 5: /* 5-element connected component */
        v = min_element(component);
        count = theta5[cardinality(ra[v] & component)]
                      [cardinality(rb[v] & component)];
        v_set = component ^ pof2(v);
        v = min_element(v_set);
        count += theta5[cardinality(ra[v] & component)]
                       [cardinality(rb[v] & component)];
        v_set ^= pof2(v);
        v = min_element(v_set);
        count += theta5[cardinality(ra[v] & component)]
                       [cardinality(rb[v] & component)];
        v_set ^= pof2(v);
        v = min_element(v_set);
        count += theta5[cardinality(ra[v] & component)]
                       [cardinality(rb[v] & component)];
        v = min_element(v_set ^ pof2(v));
        count += theta5[cardinality(ra[v] & component)]
                       [cardinality(rb[v] & component)];
        t *= lec5[count] * binomial[5][c];
        break;
      default: /* k > 5 */
        /* Choose split element which is comparable
           with the most count of others elements. */
        max_count = 0;
        split_v = 0; /* Statement added to avoid -Wall warning. */
        v_set = component;
        while (v_set) {
          v = min_element(v_set);
          v_set ^= pof2(v);
          count = cardinality((ra[v] | rb[v]) & component) + 1;
          if (count == k) {
            split_v = v;
            break;
          }
          else if (count > max_count) {
            split_v = v;
            max_count = count;
          }
        }
        m_set = component & ~(ra[split_v] | rb[split_v] | pof2(split_v));

        /* Consider all admissible partitions of m_set. */
        m_sub_set = EMPTY_SET;
        t1 = 0;

      S_LOOP_BEGIN:
        sub_component = m_sub_set | (ra[split_v] & component);
        switch (cardinality(sub_component)) {
          case 0:
          case 1:
            t2 = 1;
            break;
          case 2:
            v = min_element(sub_component);
            t2 = ((ra[v] | rb[v]) & sub_component) ? 1 : 2;
            break;
          case 3:
            v = min_element(sub_component);
            v_set = sub_component ^ pof2(v);
            w = min_element(v_set);
            t2 = lec3[cardinality(ra[v] & sub_component) +
                      cardinality(ra[w] & sub_component) +
                      cardinality(ra[min_element(v_set ^ pof2(w))] &
                                  sub_component)];
            break;
          case 4:
            v = min_element(sub_component);
            v_set = sub_component ^ pof2(v);
            w = min_element(v_set);
            v_set ^= pof2(w);
            x = min_element(v_set);
            y = min_element(v_set ^ pof2(x));
            count = theta4[cardinality(ra[v] & sub_component)] +
                    theta4[cardinality(ra[w] & sub_component)] +
                    theta4[cardinality(ra[x] & sub_component)] +
                    theta4[cardinality(ra[y] & sub_component)];
            if (count == 2 &&
                cardinality((ra[v] | ra[w] | ra[x] | ra[y]) &
                            sub_component) == 1)
              t2 = 8;
            else if (count == 5 &&
                     cardinality(((ra[v] | ra[w] | ra[x] | ra[y]) |
                                  rb[v] | rb[w] | rb[x] | rb[y]) &
                                 sub_component) != 4)
              t2 = 4;
            else
              t2 = lec4[count];
            break;
          default:
            t2 = cpcp_i(sub_component, ra, rb);
        }
        sub_component = component & ~(sub_component | pof2(split_v));
        switch (cardinality(sub_component)) {
          case 0:
          case 1:
            break;
          case 2:
            v = min_element(sub_component);
            if (((ra[v] | rb[v]) & sub_component) == EMPTY_SET)
              t2 *= 2;
            break;
          case 3:
            v = min_element(sub_component);
            v_set = sub_component ^ pof2(v);
            w = min_element(v_set);
            t2 *= lec3[cardinality(ra[v] & sub_component) +
                       cardinality(ra[w] & sub_component) +
                       cardinality(ra[min_element(v_set ^ pof2(w))] &
                                   sub_component)];
            break;
          case 4:
            v = min_element(sub_component);
            v_set = sub_component ^ pof2(v);
            w = min_element(v_set);
            v_set ^= pof2(w);
            x = min_element(v_set);
            y = min_element(v_set ^ pof2(x));
            count = theta4[cardinality(ra[v] & sub_component)] +
                    theta4[cardinality(ra[w] & sub_component)] +
                    theta4[cardinality(ra[x] & sub_component)] +
                    theta4[cardinality(ra[y] & sub_component)];
            if (count == 2 &&
                cardinality((ra[v] | ra[w] | ra[x] | ra[y]) &
                            sub_component) == 1)
              t2 *= 8;
            else if (count == 5 &&
                     cardinality(((ra[v] | ra[w] | ra[x] | ra[y]) |
                                  rb[v] | rb[w] | rb[x] | rb[y]) &
                                 sub_component) != 4)
              t2 *= 4;
            else
              t2 *= lec4[count];
            break;
          default:
            t2 *= cpcp_i(sub_component, ra, rb);
        }
        t1 += t2;

      S_TAKE_NEXT_SUBSET:
        /* If this is the last subset then ... */
        if (m_sub_set == m_set) goto S_LOOP_END;
        /* ... else calculate the next subset of m_set. */
        v_set = pof2(min_element(m_set & ~m_sub_set));
        m_sub_set = (m_sub_set | v_set) & ~(v_set - 1);
        v_set = m_set & ~m_sub_set;
        while (v_set) {
          v = min_element(v_set);
          if (rb[v] & m_sub_set) goto S_TAKE_NEXT_SUBSET;
          v_set ^= pof2(v);
        }
        goto S_LOOP_BEGIN;

      S_LOOP_END:
        t *= t1 * binomial[k][c];
    }
  }
  return t;
}
